package com.medm.devicekit.sample

import android.bluetooth.BluetoothAdapter
import android.content.Context
import android.content.Intent
import android.os.Build
import android.os.Bundle
import android.text.Html
import android.text.Spanned
import android.text.SpannedString
import android.util.Xml
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.*
import androidx.activity.result.ActivityResultLauncher
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.fragment.app.ListFragment
import com.google.android.material.datepicker.MaterialDatePicker
import com.google.android.material.dialog.MaterialAlertDialogBuilder
import com.google.android.material.snackbar.BaseTransientBottomBar
import com.google.android.material.snackbar.Snackbar
import com.google.android.material.switchmaterial.SwitchMaterial
import com.google.android.material.textfield.TextInputEditText
import com.medm.devicekit.*
import org.xmlpull.v1.XmlPullParser
import org.xmlpull.v1.XmlPullParserException
import java.io.IOException
import java.io.StringReader
import java.text.SimpleDateFormat
import java.util.*

class DataCollectionFragment(val locationServicesRequester: LocationServicesRequester) : ListFragment()
{
    private val mReadingsListAdapter = ReadingsListAdapter()

    private lateinit var mPermissionsRequest : ActivityResultLauncher<Array<String>>
    private lateinit var mTurnOnBluetooth : ActivityResultLauncher<Intent>

    override fun onCreate(savedInstanceState: Bundle?)
    {
        super.onCreate(savedInstanceState)
        listAdapter = mReadingsListAdapter

        mPermissionsRequest = registerForActivityResult(ActivityResultContracts.RequestMultiplePermissions()) { map ->
            if (map.isNotEmpty() && map.all { it.value })
                startCollectData()
        }

        mTurnOnBluetooth = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) {
            if (it.resultCode == AppCompatActivity.RESULT_OK)
                startCollectData()
        }
    }

    private val deviceList: Array<IDeviceDescription?>
        get()
        {
            try
            {
                return MedMDeviceKit.getDeviceManager().devicesList
            } catch (ex: Exception)
            {
            }
            return arrayOfNulls(0)
        }

    override fun onListItemClick(l: ListView, v: View, position: Int, id: Long)
    {
        val container = mReadingsListAdapter.getReading(position)
        container.isCollapsed = !container.isCollapsed
        mReadingsListAdapter.notifyDataSetChanged()
    }

    fun startCollectData() : CollectorStopToken
    {
        mReadingsListAdapter.clear()
        mReadingsListAdapter.notifyDataSetChanged()

        MedMDeviceKit.getCollector().setDataCallback(object : IDataCallback
        {
            override fun onNewData(source: IDeviceDescription, reading: String)
            {
                val sourceModel = source.modelName
                activity?.runOnUiThread {
                    Toast.makeText(activity, getString(R.string.new_data) + " " + sourceModel, Toast.LENGTH_SHORT).show()
                    mReadingsListAdapter.addReading(reading)
                    mReadingsListAdapter.notifyDataSetChanged()
                }
            }

            override fun onDataCollectionStopped() {}
        })
        .setDeviceStatusCallback(object : IDeviceStatusCallback
        {
            override fun onConnected(device: IDeviceDescription) {
                Toast.makeText(activity, getString(R.string.connected_to) + " " + device.modelName, Toast.LENGTH_SHORT).show()
            }

            override fun onDisconnected(device: IDeviceDescription, status: OperationStatus, statusInfo: String) {
                Toast.makeText(activity, getString(R.string.disconnected_from) + " " + device.modelName, Toast.LENGTH_SHORT).show()
            }
        })
        .setAddDeviceCallback(object : IAddDeviceCallback
        {
            override fun onSuccess(device: IDeviceDescription) {
                activity?.runOnUiThread { Toast.makeText(activity, device.modelName + " paired successfully", Toast.LENGTH_LONG).show() }
            }

            override fun onFailure(device: IDeviceDescription, operationStatus: OperationStatus, errDescription: String) {
                activity?.runOnUiThread { Toast.makeText(activity, device.address + " failed to pair", Toast.LENGTH_LONG).show() }
            }
        })
        .setErrorCallback(object : IErrorCallback
        {
            override fun onPermissionsRequired(ex: NotEnoughPermissionsException) {
                if (ex.requiredPermissions.contains(PermissionType.LocationServices))
                {
                    locationServicesRequester.requestTurnOn(object: LocationServicesRequester.Listener {
                        override fun onResult(granted: Boolean) {
                            if (granted)
                                startCollectData()
                        }
                    })
                }
                else { mPermissionsRequest.launch(ex.requiredPermissionsStrings) }
            }

            override fun onBluetoothTurnedOff(bluetoothState: BluetoothState) {
                val snackbar = Snackbar.make(requireActivity().findViewById(android.R.id.content), R.string.bluetooth_off_warning, BaseTransientBottomBar.LENGTH_INDEFINITE)

                if (MedMDeviceKit.getBluetoothState().hasPermissionsToTurnOnBluetooth(activity))
                {
                    snackbar.setAction(R.string.bluetooth_off_button) {
                        val intent = Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE)
                        mTurnOnBluetooth.launch(intent)
                    }
                }

                snackbar.show()
                bluetoothState.onBluetoothTurnedOn { snackbar.dismiss() }
            }
        })

        return try
        {
            MedMDeviceKit.getCollector().startWithPostponedAdding()
        } catch (ex: FeatureUnsupportedException)
        {
            MedMDeviceKit.getCollector().start()
        }
    }

    private inner class ReadingContainer(val readingXML: String)
    {
        private val readingShort = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) Html.fromHtml(parseXml(), Html.FROM_HTML_MODE_COMPACT) else Html.fromHtml(parseXml())
        var isCollapsed = true

        private inner class SimpleReading(private val type: String)
        {
            private var value: String? = null
            private var units: String? = null
            private var measuredAt = "UNKNOWN"
            fun setUnits(units: String?)
            {
                this.units = units
            }

            fun addValuePart(value: String)
            {
                this.value = if (this.value == null) value else this.value + " " + value
            }

            fun setMeasuredAt(measuredAt: String)
            {
                this.measuredAt = measuredAt
            }

            override fun toString(): String
            {
                val typeCap = type.substring(0, 1).toUpperCase() + type.substring(1)
                return ("<h1>" + typeCap + "</h1>"
                        + "<b>Value:</b> <i>" + value + (if (units != null) "<small> $units</small>" else "") + "</i>" + "<br/>"
                        + "<b>Time:</b> <i>" + measuredAt.trim { it <= ' ' } + "</i>")
            }

        }

        /*
            See https://health.medm.com/docs/api/v3/index.html#measurements for XML examples
            for different reading types. Note, that some tags could be missing in SDK, as they
            are server-generated.
         */
        private fun parseXml(): String
        {
            val xpp = Xml.newPullParser()
            var reading: SimpleReading? = null
            try
            {
                xpp.setInput(StringReader(readingXML))
                var eventType = xpp.eventType
                while (eventType != XmlPullParser.END_DOCUMENT)
                {
                    when (eventType)
                    {
                        XmlPullParser.START_DOCUMENT ->
                        {
                        }
                        XmlPullParser.START_TAG ->
                        {
                            val name = xpp.name
                            if (name.startsWith("measurements-"))
                            {
                                reading = SimpleReading(name.replaceFirst("measurements-".toRegex(), "").replaceFirst("-stream".toRegex(), ""))
                            } else if (reading != null)
                            {
                                if (name.equals("value", ignoreCase = true) // Glucose, weight
                                    || name.equals("systolic", ignoreCase = true) // Blood pressure
                                    || name.equals("diastolic", ignoreCase = true) // Blood pressure
                                    || name.equals("pulse", ignoreCase = true) // Blood pressure
                                    || name.equals("rate", ignoreCase = true) // Respiration rate
                                    || name.equals("steps", ignoreCase = true) // Activity
                                    || name.equals("channel", ignoreCase = true) //channel names
                                    || name.equals("sleep-time", ignoreCase = true) // Sleep
                                    || name.equals("heartrate", ignoreCase = true) // Heart rate JSON chunk
                                    || name.equals("oxygen", ignoreCase = true) // SpO2 JSON chunk
                                    || name.equals("temperature", ignoreCase = true) // Temperature JSON chunk
                                    || name.equals("spirometry", ignoreCase = true) // Spirometry JSON chunk
                                    || name.equals("body", ignoreCase = true) // Note body
                                    || name.equals("inr", ignoreCase = true) // Blood coagulation
                                )
                                {
                                    reading.addValuePart(xpp.nextText().trim { it <= ' ' })
                                } else if (name.equals("units", ignoreCase = true))
                                {
                                    reading.setUnits(xpp.nextText().trim { it <= ' ' })
                                } else if (name.equals("measured-at-local", ignoreCase = true))
                                {
                                    reading.setMeasuredAt(xpp.nextText())
                                }
                            }
                        }
                        XmlPullParser.END_TAG ->
                        {
                        }
                    }
                    eventType = xpp.next()
                }
            } catch (e: XmlPullParserException)
            {
                e.printStackTrace()
            } catch (e: IOException)
            {
                e.printStackTrace()
            }
            return reading?.toString() ?: "Error parsing"
        }

        override fun equals(other: Any?): Boolean
        {
            return other is ReadingContainer && readingXML == other.readingXML
        }

        override fun hashCode(): Int
        {
            return readingXML.hashCode()
        }

        val reading: Spanned
            get() = if (isCollapsed) readingShort else SpannedString(readingXML)
    }

    private inner class ReadingsListAdapter : BaseAdapter()
    {
        private val mReadings = ArrayList<ReadingContainer>()

        fun addReading(reading: String)
        {
            if (!mReadings.contains(ReadingContainer(reading)))
                mReadings.add(ReadingContainer(reading))
        }

        fun getReading(position: Int): ReadingContainer { return mReadings[position] }

        fun clear() { mReadings.clear() }

        override fun getCount(): Int { return mReadings.size }

        override fun getItem(i: Int): Any { return mReadings[i] }

        override fun getItemId(i: Int): Long { return i.toLong() }

        override fun getView(i: Int, view: View?, viewGroup: ViewGroup): View
        {
            val viewHolder : ViewHolder
            val container = mReadings[i]
            var readingView = view

            if (readingView == null)
            {
                readingView = layoutInflater.inflate(R.layout.reading, null)
                viewHolder = ViewHolder()
                viewHolder.reading = readingView.findViewById(R.id.reading)
                readingView.tag = viewHolder
            } else
            {
                viewHolder = readingView.tag as ViewHolder
            }

            viewHolder.reading?.text = container.reading

            return readingView!!
        }
    }

    internal class ViewHolder { var reading: TextView? = null }
}