package com.medm.devicekit.sample

import android.bluetooth.BluetoothAdapter
import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.os.Parcelable
import android.util.Log
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.BaseAdapter
import android.widget.ListView
import android.widget.TextView
import android.widget.Toast
import androidx.activity.result.ActivityResultLauncher
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.core.view.isVisible
import androidx.fragment.app.ListFragment
import com.google.android.material.dialog.MaterialAlertDialogBuilder
import com.google.android.material.snackbar.BaseTransientBottomBar
import com.google.android.material.snackbar.Snackbar
import com.medm.devicekit.*
import java.util.ArrayList

class DeviceScanFragment(val locationServicesRequester: LocationServicesRequester) : ListFragment()
{
    private val TAG = DeviceScanFragment::class.java.simpleName

    private val mDeviceListAdapter = DeviceListAdapter()
    private var mDeviceAddStopToken: DeviceAddingCancellationToken? = null
    private var mListEnabled = true // Prevent adding two devices at once
    var mProgressDialog: AlertDialog? = null
    private var mDeviceToAdd: IDeviceDescription? = null
    private lateinit var mAddDevice : ActivityResultLauncher<Intent>
    private lateinit var mRequestPermissions : ActivityResultLauncher<Array<String>>
    private lateinit var mTurnOnBluetooth : ActivityResultLauncher<Intent>

    override fun onCreate(savedInstanceState: Bundle?)
    {
        super.onCreate(savedInstanceState)
        listAdapter = mDeviceListAdapter

        mAddDevice = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) {
            if (it.resultCode == AppCompatActivity.RESULT_OK)
                addDevice(mDeviceToAdd!!)
            else
            {
                mDeviceListAdapter.notifyDataSetChanged()
                startScan()
            }
        }

        mRequestPermissions = registerForActivityResult(ActivityResultContracts.RequestMultiplePermissions()) { map ->
            if (map.isNotEmpty() && map.all { it.value })
                startScan()
        }

        mTurnOnBluetooth = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) {
            if (it.resultCode == AppCompatActivity.RESULT_OK)
                startScan()
        }
    }

    override fun onListItemClick(l: ListView, v: View, position: Int, id: Long)
    {
        super.onListItemClick(l, v, position, id)

        val device = mDeviceListAdapter.getDevice(position)

        if (device.settings.isNotEmpty())
        {
            val intent = Intent(activity, DeviceSettingsActivity::class.java)
            intent.putExtra(DeviceSettingsActivity.EXTRA_DEVICE, device as Parcelable)
            intent.putExtra(DeviceSettingsActivity.EXTRA_BEFORE_ADDING, true)
            mAddDevice.launch(intent)
            mDeviceToAdd = device
        }
        else
            addDevice(device)
    }

    fun startScan() : ScannerStopToken
    {
        return MedMDeviceKit.getScanner().setCallback(object : IScannerCallback
        {
            override fun onDeviceFound(device: IDeviceDescription)
            {
                Log.d(TAG, "onKnownDeviceFound address=" + device.address + " name='" + device.modelName + "'")
                if (!mDeviceListAdapter.containsDevice(device))
                {
                    activity?.runOnUiThread {
                        mDeviceListAdapter.addDevice(device)
                        mDeviceListAdapter.notifyDataSetChanged()
                    }
                }
            }

            override fun onAmbiguousDeviceFound(iDeviceDescriptions: Array<IDeviceDescription>)
            {
                Log.d(TAG, "onAmbiguousDeviceFound alternatives number=" + iDeviceDescriptions.size)

                activity?.runOnUiThread {
                    mDeviceListAdapter.addAmbiguousDevice(iDeviceDescriptions)
                    mDeviceListAdapter.notifyDataSetChanged()
                }
            }

            override fun onScanFinished()
            {
                Log.d(TAG, "onScanFinished")
            }
        })
        .setErrorCallback(object : IErrorCallback
        {
            override fun onPermissionsRequired(ex: NotEnoughPermissionsException) {
                if (ex.requiredPermissions.contains(PermissionType.LocationServices)) {
                    locationServicesRequester.requestTurnOn(object : LocationServicesRequester.Listener {
                        override fun onResult(granted: Boolean) {
                            if (granted)
                                startScan()
                        }
                    })
                } else {
                    mRequestPermissions.launch(ex.requiredPermissionsStrings)
                }
            }

            override fun onBluetoothTurnedOff(bluetoothState: BluetoothState) {
                val snackbar = Snackbar.make(requireActivity().findViewById(android.R.id.content), R.string.bluetooth_off_warning, BaseTransientBottomBar.LENGTH_INDEFINITE)

                if (MedMDeviceKit.getBluetoothState().hasPermissionsToTurnOnBluetooth(activity))
                {
                    snackbar.setAction(R.string.bluetooth_off_button) {
                        val intent = Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE)
                        mTurnOnBluetooth.launch(intent)
                    }
                }

                snackbar.show()
                bluetoothState.onBluetoothTurnedOn { snackbar.dismiss() }
            }
        }).start()
    }

    private fun addDevice(device: IDeviceDescription)
    {
        Log.d(TAG, "addDevice address=" + device.address + " name='" + device.modelName + "'")

        mListEnabled = false

        mProgressDialog = MaterialAlertDialogBuilder(requireActivity()).setView(R.layout.device_add_progress).create()
        mProgressDialog!!.setCancelable(true)
        mProgressDialog!!.setCanceledOnTouchOutside(true)
        mProgressDialog!!.setOnCancelListener { mDeviceAddStopToken?.cancel() }
        mProgressDialog!!.show()
        mProgressDialog!!.findViewById<TextView>(R.id.message)?.text = getString(R.string.adding_device, device.name)

        val callback: IAddDeviceCallback = object : IAddDeviceCallback
        {
            override fun onSuccess(source: IDeviceDescription)
            {
                mListEnabled = true
                activity?.runOnUiThread {
                    mProgressDialog?.dismiss()
                    Toast.makeText(activity, getString(R.string.add_success) + " " + source.modelName, Toast.LENGTH_SHORT).show()
                }
                mDeviceAddStopToken = null
            }

            override fun onFailure(source: IDeviceDescription, status: OperationStatus, statusDescription: String)
            {
                mListEnabled = true
                activity?.runOnUiThread {
                    mProgressDialog?.dismiss()
                    Toast.makeText(activity, getString(R.string.add_fail) + " " + source.modelName, Toast.LENGTH_SHORT).show()
                }
                mDeviceAddStopToken = null
            }
        }

        mDeviceAddStopToken = MedMDeviceKit.getDeviceManager().addDevice(device, callback)
    }

    // Adapter for holding devices found through scanning.
    private inner class DeviceListAdapter : BaseAdapter()
    {
        private val mDevices = ArrayList<IDeviceDescription>()

        fun addDevice(device: IDeviceDescription)
        {
            if (!containsDevice(device)) mDevices.add(device)
        }

        fun containsDevice(device: IDeviceDescription): Boolean
        {
            for (savedDevice in mDevices)
            {
                if (savedDevice.address == device.address && savedDevice.sku == device.sku)
                {
                    return true
                }
            }
            return false
        }

        fun addAmbiguousDevice(deviceVariants: Array<IDeviceDescription>)
        {
            for (device in deviceVariants)
            {
                addDevice(device)
            }
        }

        fun getDevice(position: Int): IDeviceDescription { return mDevices[position] }

        fun clear() { mDevices.clear() }

        override fun getCount(): Int { return mDevices.size }

        override fun getItem(i: Int): Any {  return mDevices[i] }

        override fun getItemId(i: Int): Long { return i.toLong() }

        override fun isEnabled(position: Int): Boolean { return mListEnabled }

        override fun getView(i: Int, view: View?, viewGroup: ViewGroup): View
        {
            var deviceView = view
            val viewHolder: ViewHolder

            // General ListView optimization code.
            if (deviceView == null)
            {
                deviceView = layoutInflater.inflate(R.layout.listitem_device, null)
                viewHolder = ViewHolder()
                viewHolder.deviceAddress = deviceView.findViewById(R.id.device_address)
                viewHolder.deviceName = deviceView.findViewById(R.id.device_name)
                viewHolder.deviceMeasurements = deviceView.findViewById(R.id.device_measurements)
                deviceView.tag = viewHolder
            } else
            {
                viewHolder = deviceView.tag as ViewHolder
            }

            val device = mDevices[i]
            val deviceName = device.modelName
            if (deviceName.isNotEmpty()) viewHolder.deviceName!!.text = deviceName else viewHolder.deviceName!!.setText(R.string.unknown_device)
            viewHolder.deviceAddress!!.text = device.address
            viewHolder.deviceMeasurements!!.text = AddedDevicesFragment.formatMeasurementTypes(device.measurementTypes)

            return deviceView!!
        }
    }

    internal class ViewHolder
    {
        var deviceName: TextView? = null
        var deviceAddress: TextView? = null
        var deviceMeasurements: TextView? = null
    }
}