/*
 * Copyright (C) 2020 MedM Global
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.medm.devicekit.sample

import android.os.Bundle
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.*
import androidx.appcompat.app.AppCompatActivity
import androidx.core.widget.addTextChangedListener
import androidx.core.widget.doAfterTextChanged
import androidx.fragment.app.FragmentActivity
import androidx.fragment.app.ListFragment
import com.google.android.material.radiobutton.MaterialRadioButton
import com.google.android.material.textfield.TextInputLayout
import com.medm.devicekit.DeviceSetting
import com.medm.devicekit.DeviceSetting.Kind
import com.medm.devicekit.IDeviceDescription

class DeviceSettingsActivity : AppCompatActivity()
{
    private val TAG = DeviceSettingsActivity::class.java.simpleName
    private var mBeforeAdding = false
    private lateinit var settingsListAdapter: SettingsListAdapter

    override fun onCreate(savedInstanceState: Bundle?)
    {
        super.onCreate(savedInstanceState)
        supportActionBar!!.setTitle(R.string.title_settings)
        setContentView(R.layout.settings)
        val device  = intent.getParcelableExtra<IDeviceDescription>(EXTRA_DEVICE)
        mBeforeAdding = intent.getBooleanExtra(EXTRA_BEFORE_ADDING, false)
        val settingsListFragment = supportFragmentManager.findFragmentById(R.id.list_fragment) as ListFragment

        if (device != null) {
            settingsListAdapter = SettingsListAdapter(device)
            settingsListFragment.listAdapter = settingsListAdapter
        }
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean
    {
        menuInflater.inflate(R.menu.settings, menu)
        menu.findItem(R.id.menu_add_device).isVisible = mBeforeAdding
        return true
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean
    {
        when (item.itemId)
        {
            R.id.menu_add_device ->
            {
                setResult(RESULT_OK)
                finish()
                return true
            }
        }
        return super.onOptionsItemSelected(item)
    }

    private inner class SettingsListAdapter(val device: IDeviceDescription) : BaseAdapter()
    {
        var settings = device.settings

        fun getSetting(position: Int): DeviceSetting { return settings[position] }

        override fun getCount(): Int { return settings.size }

        override fun getItem(i: Int): Any { return settings[i] }

        override fun getItemId(i: Int): Long { return i.toLong() }

        override fun getView(i: Int, view: View?, viewGroup: ViewGroup): View
        {
            // General ListView optimization code.
            if (view != null)
                return view

            val setting = settings[i]
            return if (setting.kind == Kind.EncryptionKey) fromStringSetting(setting) else fromRadioSetting(setting)
        }

        private fun fromRadioSetting(setting: DeviceSetting) : View
        {
            val viewHolder = ViewHolder()
            val settingView = layoutInflater.inflate(R.layout.device_setting, null)
            viewHolder.settingTitle = settingView.findViewById(R.id.setting_title)
            viewHolder.settingsGroup = settingView.findViewById(R.id.setting_radiogroup)
            settingView.tag = viewHolder

            viewHolder.settingTitle?.text = settingTitleFromKind(setting.kind)

            for (index: Int in setting.possibleValues.indices)
            {
                val button = MaterialRadioButton(this@DeviceSettingsActivity)
                button.text = setting.possibleValues[index]
                button.textSize = 17.0f
                button.id = index + 1
                button.isEnabled = !setting.isReadonly
                viewHolder.settingsGroup?.addView(button)
            }

            viewHolder.settingsGroup?.check(setting.valueIndex + 1)
            viewHolder.settingsGroup?.setOnCheckedChangeListener { _, checkedId ->
                setting.selectValueWithIndex(checkedId - 1)
                settings = device.settings
                runOnUiThread {
                    settingsListAdapter.notifyDataSetChanged()
                }
            }

            return settingView
        }

        private fun fromStringSetting(setting: DeviceSetting) : View
        {
            val viewHolder = StringViewHolder()
            val settingView = layoutInflater.inflate(R.layout.string_device_setting, null)
            viewHolder.settingLayout = settingView.findViewById(R.id.setting_layout)
            viewHolder.settingsValue = settingView.findViewById(R.id.setting_value)
            settingView.tag = viewHolder

            viewHolder.settingLayout?.hint = settingTitleFromKind(setting.kind)
            viewHolder.settingsValue?.setText(setting.value)
            viewHolder.settingsValue?.textSize = 17.0f
            viewHolder.settingsValue?.isEnabled = !setting.isReadonly
            viewHolder.settingsValue?.doAfterTextChanged { setting.value = it.toString() }

            return settingView
        }
    }

    internal class ViewHolder
    {
        var settingTitle: TextView? = null
        var settingsGroup: RadioGroup? = null
        var settingsValue: EditText? = null
    }

    internal class StringViewHolder
    {
        var settingLayout: TextInputLayout? = null
        var settingsValue: EditText? = null
    }

    fun settingTitleFromKind(kind: Kind) : String
    {
        val id = when (kind)
        {
            Kind.StreamOrSpot-> R.string.setting_stream_spot
            Kind.HistoryOrRealtime-> R.string.setting_history_realtime
            Kind.MeasurementLength-> R.string.setting_measurement_length
            Kind.UserNumber-> R.string.setting_user_number
            Kind.EraseDataOnAdding-> R.string.setting_erase_data_on_adding
            Kind.HeartRateMonitoring-> R.string.setting_hr_monitoring
            Kind.SpirometryTurbineType-> R.string.setting_spirometry_turbine_type
            Kind.SpirometryTestType-> R.string.setting_spirometry_test_type
            Kind.KeepHistory-> R.string.setting_keep_history
            Kind.DeviceManufacturer-> R.string.setting_device_manufacturer
            Kind.IsUserAthlete-> R.string.setting_is_athlete
            Kind.EncryptionKey-> R.string.setting_encryption_key
        }

        return getString(id)
    }

    companion object
    {
        const val EXTRA_DEVICE = "EXTRA_DEVICE"
        const val EXTRA_BEFORE_ADDING = "EXTRA_BEFORE_ADDING"
    }
}